import asyncio
from signal import SIGTERM, SIGINT

from buz.event.strategies.execution_strategy.async_execution_strategy import AsyncExecutionStrategy


class AsyncWorker:
    def __init__(
        self,
        execution_strategy: AsyncExecutionStrategy,
    ) -> None:
        self.__execution_strategy = execution_strategy

    async def start(self) -> None:
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(SIGINT, lambda: asyncio.create_task(self.stop()))
        loop.add_signal_handler(SIGTERM, lambda: asyncio.create_task(self.stop()))
        print("Starting buz worker...")
        await self.__execution_strategy.start()
        print("Buz worker stopped gracefully")

    async def stop(self) -> None:
        print("Stopping buz worker...")
        await self.__execution_strategy.stop()



import time
import asyncio
from asyncio import Future, create_task, gather, Semaphore, Event as AsyncIOEvent, sleep
from datetime import timedelta, datetime
from logging import Logger
from threading import Thread
import traceback
from typing import Optional, Type, TypeVar
from queue import Queue

from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor

from buz.event import Event, Subscriber
from buz.event.async_consumer import AsyncConsumer
from buz.event.domain.queue.queue_repository import QueueRepository
from buz.event.infrastructure.buz_kafka.consume_strategy.consume_strategy import KafkaConsumeStrategy
from buz.event.infrastructure.buz_kafka.consume_strategy.kafka_on_fail_strategy import KafkaOnFailStrategy
from buz.event.infrastructure.buz_kafka.in_memory_queue_repository import InMemoryQueueRepository
from buz.event.middleware.consume_middleware import ConsumeMiddleware
from buz.event.middleware.consume_middleware_chain_resolver import ConsumeMiddlewareChainResolver
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback
from buz.kafka import (
    KafkaConnectionConfig,
    KafkaConsumerRecord,
    ConsumerInitialOffsetPosition,
)
from buz.kafka.infrastructure.deserializers.bytes_to_message_deserializer import BytesToMessageDeserializer
from buz.kafka.infrastructure.deserializers.implementations.json_bytes_to_message_deserializer import (
    JSONBytesToMessageDeserializer,
)
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_multi_threaded_consumer_factory import (
    KafkaPythonMultiThreadedConsumerFactory,
)
from buz.kafka.infrastructure.kafka_python.kafka_poll_record import KafkaPollRecord
from buz.kafka.infrastructure.kafka_python.kafka_python_multi_threaded_consumer import KafkaPythonMultiThreadedConsumer

T = TypeVar("T", bound=Event)
KafkaConsumer = KafkaPythonMultiThreadedConsumer  # TODO: remove this alias once we comply with the interface

ConsumingTask = tuple[KafkaPollRecord, KafkaConsumer]


class BuzKafkaMultiThreadedConsumer(AsyncConsumer):
    __MAX_NUMBER_OF_POLLING_TASKS = 20
    __TIME_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE = 0.5
    __TIME_BETWEEN_COMMITS_POLLING = 0.1
    __TIME_BETWEEN_RETRIES = 5

    __commit_task_future: Optional[Future]

    def __init__(
        self,
        *,
        connection_config: KafkaConnectionConfig,
        consume_strategy: KafkaConsumeStrategy,
        on_fail_strategy: KafkaOnFailStrategy,
        queue_repository: QueueRepository[ConsumingTask],
        kafka_partition_assignors: tuple[Type[AbstractPartitionAssignor], ...] = (),
        subscribers: list[Subscriber],
        logger: Logger,
        consumer_initial_offset_position: ConsumerInitialOffsetPosition,
        deserializers_per_subscriber: dict[Subscriber, BytesToMessageDeserializer[T]],
        consume_middlewares: Optional[list[ConsumeMiddleware]] = None,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
        max_queue_size: int,
        max_records_retrieved_per_poll: int,
        min_time_between_polls_in_ms: int,
    ):
        self.__connection_config = connection_config
        self.__consume_strategy = consume_strategy
        self.__on_fail_strategy = on_fail_strategy
        self.__queue_repository = queue_repository
        self.__commit_repository: QueueRepository[ConsumingTask] = InMemoryQueueRepository(queue=Queue())
        self.__kafka_partition_assignors = kafka_partition_assignors
        self.__subscribers = subscribers
        self.__logger = logger
        self.__consumer_initial_offset_position = consumer_initial_offset_position
        self.__deserializers_per_subscriber = deserializers_per_subscriber
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self.__max_records_retrieved_per_poll = 1
        self.__subscriber_per_consumer_mapper: dict[KafkaConsumer, Subscriber] = {}
        self.__should_stop = AsyncIOEvent()
        self.__stop_commit_task = AsyncIOEvent()
        self.__stop_consumption_thread = AsyncIOEvent()
        self.__start_kafka_consumers_elapsed_time: Optional[timedelta] = None
        self.__initial_coroutines_created_elapsed_time: Optional[timedelta] = None
        self.__events_processed: int = 0
        self.__events_processed_elapsed_time: timedelta = timedelta()
        self.__polling_tasks_semaphore = Semaphore(self.__MAX_NUMBER_OF_POLLING_TASKS)
        self.__consumption_thread = Thread(target=lambda: asyncio.run(self.__consume_events_thread()))

    async def run(self) -> None:
        await self.__generate_kafka_consumers()

        start_time = datetime.now()

        self.__consumption_thread.start()
        self.__commit_task_future = create_task(self.__commit_task())
        await self.__polling_task()

        self.__events_processed_elapsed_time = datetime.now() - start_time

        await self.__perform_graceful_stop()

    async def __generate_kafka_consumers(self):
        start_time = datetime.now()
        tasks = [self.__create_kafka_consumer_for_subscriber(subscriber) for subscriber in self.__subscribers]
        await gather(*tasks)
        self.__start_kafka_consumers_elapsed_time = datetime.now() - start_time

    async def __create_kafka_consumer_for_subscriber(self, subscriber: Subscriber) -> None:
        byte_deserializer = self.__deserializers_per_subscriber.get(subscriber)
        kafka_python_consumer_factory = KafkaPythonMultiThreadedConsumerFactory(
            consumer_group=self.__consume_strategy.get_subscription_group(subscriber),
            topics=self.__consume_strategy.get_topics(subscriber),
            kafka_connection_config=self.__connection_config,
            initial_offset_position=self.__consumer_initial_offset_position,
            byte_deserializer=byte_deserializer or JSONBytesToMessageDeserializer(event_class=subscriber.handles()),  # type: ignore[arg-type]
            kafka_partition_assignors=self.__kafka_partition_assignors,
            logger=self.__logger,
        )
        kafka_consumer = kafka_python_consumer_factory.build()
        self.__subscriber_per_consumer_mapper[kafka_consumer] = subscriber
        await kafka_consumer.init()

    async def __polling_task(self) -> None:
        try:
            while not self.__should_stop.is_set():
                if not self.__queue_repository.is_empty():
                    await sleep(self.__TIME_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE)
                    continue
                raw_consuming_tasks = await gather(
                    *[
                        self.__polling_consuming_tasks(kafka_consumer=consumer)
                        for consumer, subscriber in self.__subscriber_per_consumer_mapper.items()
                    ]
                )
                poll_results = [
                    consuming_task for consuming_tasks in raw_consuming_tasks for consuming_task in consuming_tasks
                ]

                if len(poll_results) == 0:
                    await sleep(self.__TIME_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE)

                for poll_result in poll_results:
                    self.__queue_repository.push(poll_result)

        except Exception:
            self.__logger.error(f"Polling task failed with exception: {traceback.format_exc()}")
            self.__should_stop.set()

        return

    async def __commit_task(self) -> None:
        while not self.__stop_commit_task.is_set():
            await sleep(self.__TIME_BETWEEN_COMMITS_POLLING)
            await self.__commit_pending_tasks()

    async def __commit_pending_tasks(self) -> None:
        while self.__commit_repository.is_empty() is False:
            commit_record, consumer = self.__commit_repository.pop()
            await consumer.commit_poll_record(commit_record)

    async def __polling_consuming_tasks(self, kafka_consumer: KafkaConsumer) -> list[ConsumingTask]:
        async with self.__polling_tasks_semaphore:
            results = await kafka_consumer.poll(
                number_of_messages_to_poll=self.__max_records_retrieved_per_poll,
            )
        return [(result, kafka_consumer) for result in results]

    async def __consume_events_thread(self) -> None:
        while not self.__stop_consumption_thread.is_set():
            if self.__queue_repository.is_empty():
                await sleep(self.__TIME_BETWEEN_POLLS_IF_THERE_ARE_TASKS_IN_THE_QUEUE)
                continue

            kafka_poll_record, consumer = self.__queue_repository.pop()
            subscriber = self.__subscriber_per_consumer_mapper[consumer]

            try:
                await consumer.consume(
                    kafka_poll_record=kafka_poll_record,
                    consumption_callback=lambda kafka_record: self.__consumption_callback(subscriber, kafka_record),
                )
                self.__commit_repository.push((kafka_poll_record, consumer))
            except Exception as ex:
                self.__should_stop.set()
                consumer.reject_poll_record(kafka_poll_record)
                raise ex

    def __consumption_callback(self, subscriber: Subscriber, message: KafkaConsumerRecord[T]) -> None:
        self.__consume_middleware_chain_resolver.resolve(
            event=message.value, subscriber=subscriber, consume=self.__perform_consume
        )

    def __perform_consume(self, event: T, subscriber: Subscriber) -> None:
        should_retry = True
        while should_retry is True:
            try:
                self.__events_processed += 1
                return subscriber.consume(event)
            except Exception as exc:
                self.__events_processed -= 1
                self.__logger.warning(f"Event {event.id} could not be consumed by the subscriber {subscriber.fqn}")

                if self.__should_retry(event, subscriber) is True:
                    self.__register_retry(event, subscriber)
                    time.sleep(self.__TIME_BETWEEN_RETRIES)
                    continue

                if self.__on_fail_strategy == KafkaOnFailStrategy.STOP_ON_FAIL:
                    raise exc

                if self.__reject_callback:
                    self.__reject_callback.on_reject(event=event, subscribers=[subscriber])

                return

    def __should_retry(self, event: Event, subscriber: Subscriber) -> bool:
        if self.__consume_retrier is None:
            return False

        return self.__consume_retrier.should_retry(event, [subscriber])

    def __register_retry(self, event: Event, subscriber: Subscriber) -> None:
        if self.__consume_retrier is None:
            return None

        return self.__consume_retrier.register_retry(event, [subscriber])

    async def stop(self) -> None:
        self.__should_stop.set()
        self.__logger.info("Worker stop requested. Waiting for finalize the current task")

    async def __perform_graceful_stop(self) -> None:
        self.__logger.info("Stopping consuming thread....")
        await self.__manage_comsuption_thread_stopping()

        self.__logger.info("Commiting pending tasks")
        await self.__manage_commiting_task_stopping()

        self.__logger.info("Stopping kafka consumers...")
        await self.__manage_kafka_consumers_stopping()

        self.__logger.info("All kafka consumers stopped")
        self.__print_statistics()

    async def __manage_comsuption_thread_stopping(self) -> None:
        self.__stop_consumption_thread.set()
        await asyncio.to_thread(self.__consumption_thread.join)

    async def __manage_commiting_task_stopping(self) -> None:
        if self.__commit_task_future:
            self.__stop_commit_task.set()
            await self.__commit_task_future
        await self.__commit_pending_tasks()

    async def __manage_kafka_consumers_stopping(self) -> None:
        for kafka_consumer in self.__subscriber_per_consumer_mapper.keys():
            await kafka_consumer.stop()

    def __print_statistics(self) -> None:
        self.__logger.info("Number of subscribers: %d", len(self.__subscribers))
        self.__logger.info(f"Start kafka consumers elapsed time: {self.__start_kafka_consumers_elapsed_time}")
        self.__logger.info(f"Initial coroutines created elapsed time: {self.__initial_coroutines_created_elapsed_time}")
        self.__logger.info(f"Events processed: {self.__events_processed}")
        self.__logger.info(f"Events processed elapsed time: {self.__events_processed_elapsed_time}")



from __future__ import annotations

from logging import Logger
from ssl import SSLContext
from typing import Callable, Optional, cast

from aiokafka import AIOKafkaConsumer, TopicPartition, OffsetAndMetadata
from aiokafka.helpers import create_ssl_context
from asgiref.sync import sync_to_async

from buz.kafka.domain.exceptions.not_valid_kafka_message_exception import NotValidKafkaMessageException
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.infrastructure.deserializers.byte_deserializer import ByteDeserializer
from buz.kafka.infrastructure.kafka_python.kafka_poll_record import KafkaPollRecord
from buz.kafka.infrastructure.kafka_python.rebalance_ready import RebalanceReady
from buz.kafka.infrastructure.kafka_python.simple_kafka_lock_rebalancer import SimpleKafkaLockRebalancer
from buz.kafka.infrastructure.kafka_python.translators.consumer_initial_offset_position_translator import (
    ConsumerInitialOffsetPositionTranslator,
)
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer


class KafkaPythonMultiThreadedConsumer(RebalanceReady):
    __DEFAULT_POLL_TIMEOUT_MS = 0
    __DEFAULT_SESSION_TIMEOUT_MS = 1000 * 60
    __DEFAULT_HEARTBEAT_INTERVAL_MS = 1000 * 15
    __DEFAULT_MAX_POLL_INTERVAL = 2147483647

    def __init__(
        self,
        *,
        consumer_group: str,
        topics: list[str],
        connection_config: KafkaConnectionConfig,
        initial_offset_position: ConsumerInitialOffsetPosition,
        byte_deserializer: ByteDeserializer,
        header_serializer: KafkaHeaderSerializer,
        partition_assignors: tuple,
        logger: Logger,
        session_timeout_ms: int = __DEFAULT_SESSION_TIMEOUT_MS,
    ) -> None:
        self.__consumer_group = consumer_group
        self.__topics = topics
        self.__initial_offset_position = initial_offset_position
        self.__connection_config = connection_config
        self.__byte_deserializer = byte_deserializer
        self.__header_serializer = header_serializer
        self.__partition_assignors = partition_assignors
        self.__logger = logger
        self.__session_timeout_ms = session_timeout_ms
        self.__pending_messages: int = 0
        self.__consumer = self.__generate_consumer()

    def __generate_consumer(self) -> AIOKafkaConsumer:
        sasl_mechanism: Optional[str] = None
        ssl_context: Optional[SSLContext] = None

        if self.__connection_config.credentials.sasl_mechanism is not None:
            sasl_mechanism = self.__connection_config.credentials.sasl_mechanism.value

        if self.__connection_config.credentials.security_protocol == KafkaSupportedSecurityProtocols.SASL_SSL:
            ssl_context = create_ssl_context()

        consumer = AIOKafkaConsumer(
            None,
            ssl_context=ssl_context,
            bootstrap_servers=self.__connection_config.bootstrap_servers,
            security_protocol=self.__connection_config.credentials.security_protocol.value,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_username=self.__connection_config.credentials.user,
            sasl_plain_password=self.__connection_config.credentials.password,
            client_id=self.__connection_config.client_id,
            group_id=self.__consumer_group,
            enable_auto_commit=False,
            auto_offset_reset=ConsumerInitialOffsetPositionTranslator.to_kafka_supported_format(
                self.__initial_offset_position
            ),
            session_timeout_ms=self.__session_timeout_ms,
            heartbeat_interval_ms=self.__DEFAULT_HEARTBEAT_INTERVAL_MS,
            partition_assignment_strategy=list(self.__partition_assignors),
            max_poll_interval_ms=self.__DEFAULT_MAX_POLL_INTERVAL,
            rebalance_timeout_ms=self.__DEFAULT_MAX_POLL_INTERVAL,
        )

        consumer.subscribe(topics=self.__topics, listener=SimpleKafkaLockRebalancer(self))
        return consumer

    async def init(self) -> None:
        await self.__consumer.start()

    async def poll(
        self,
        *,
        timeout_ms: int = __DEFAULT_POLL_TIMEOUT_MS,
        number_of_messages_to_poll: Optional[int] = None,
    ) -> list[KafkaPollRecord]:
        poll_results = await self.__consumer.getmany(
            timeout_ms=timeout_ms,
            max_records=number_of_messages_to_poll,
        )

        results = [
            cast(KafkaPollRecord, consumer_record)
            for consumer_records in poll_results.values()
            for consumer_record in consumer_records
        ]

        self.__pending_messages += len(results)

        return results

    async def consume(
        self,
        *,
        kafka_poll_record: KafkaPollRecord,
        consumption_callback: Callable[[KafkaConsumerRecord], None],
    ) -> None:
        try:
            if kafka_poll_record.value is None:
                raise NotValidKafkaMessageException("Message is None")

            await sync_to_async(
                lambda: consumption_callback(
                    KafkaConsumerRecord(
                        value=self.__byte_deserializer.deserialize(kafka_poll_record.value),
                        headers=self.__header_serializer.deserialize(kafka_poll_record.headers),
                    )
                ),
                thread_sensitive=True,
            )()

        except NotValidKafkaMessageException:
            self.__logger.error(
                f'The message "{str(kafka_poll_record.value)}" is not valid, it will be consumed but not processed'
            )

    async def commit_poll_record(self, poll_record: KafkaPollRecord) -> None:
        offset = {
            TopicPartition(topic=poll_record.topic, partition=poll_record.partition): OffsetAndMetadata(
                poll_record.offset + 1, ""
            )
        }
        await self.__consumer.commit(offset)
        self.__pending_messages -= 1

    def reject_poll_record(self, poll_record: KafkaPollRecord) -> None:
        self.__logger.warning(f"Message {str(poll_record.value)} will be rejected")
        self.__pending_messages -= 1

    async def stop(self) -> None:
        self.__logger.info(f"Closing connection of consumer with group_id={self.__consumer_group}")
        self.__pending_messages = 0
        await self.__consumer.stop()

    def is_ready_for_rebalancing(self) -> bool:
        return self.__pending_messages == 0



from __future__ import annotations

import re

from kafka import KafkaClient
from kafka.admin import KafkaAdminClient as KafkaPythonLibraryAdminClient, NewTopic
from kafka.errors import TopicAlreadyExistsError

from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_topic import KafkaTopic
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient

INTERNAL_KAFKA_TOPICS = {"__consumer_offsets", "_schema"}


class KafkaPythonAdminClient(KafkaAdminClient):
    __PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE = 36

    def __init__(self, *, config: KafkaConnectionConfig):
        self._config = config
        self._config_in_library_format = self.__get_kafka_config_in_library_format(config)
        self._kafka_admin = KafkaPythonLibraryAdminClient(**self._config_in_library_format)
        self._kafka_client = KafkaClient(**self._config_in_library_format)

    def __get_kafka_config_in_library_format(self, config: KafkaConnectionConfig) -> dict:
        return {
            "client_id": config.client_id,
            "bootstrap_servers": config.bootstrap_servers,
            "security_protocol": config.credentials.security_protocol.value,
            "sasl_mechanism": config.credentials.sasl_mechanism,
            "sasl_plain_username": config.credentials.user,
            "sasl_plain_password": config.credentials.password,
        }

    def create_topics(
        self,
        *,
        topics: set[KafkaTopic],
    ) -> None:
        new_topics = [
            NewTopic(
                name=topic.name,
                num_partitions=topic.partitions,
                replication_factor=topic.replication_factor,
            )
            for topic in topics
        ]

        try:
            self._kafka_admin.create_topics(new_topics=new_topics)
        except TopicAlreadyExistsError as error:
            topic_names = self.__get_list_of_kafka_topics_from_topic_already_exists_error(error)
            raise KafkaTopicsAlreadyCreatedException(topic_names=topic_names)

    def __get_list_of_kafka_topics_from_topic_already_exists_error(self, error: TopicAlreadyExistsError) -> list[str]:
        message = str(error)
        response_message = re.search(r"topic_errors=\[.*?]", message)
        topic_messages = re.findall(
            r"topic='[^']*', error_code=" + str(self.__PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE), response_message[0]  # type: ignore
        )

        return [re.search("'.*'", topic_message)[0].strip("'") for topic_message in topic_messages]  # type: ignore

    def get_topics(
        self,
    ) -> set[str]:
        return set(self._kafka_admin.list_topics()) - INTERNAL_KAFKA_TOPICS

    def delete_topics(
        self,
        *,
        topics: set[str],
    ) -> None:
        self._kafka_admin.delete_topics(
            topics=topics,
        )

    def _wait_for_cluster_update(self) -> None:
        future = self._kafka_client.cluster.request_update()
        self._kafka_client.poll(future=future)



from aiokafka import ConsumerRebalanceListener
import asyncio

from buz.kafka.infrastructure.kafka_python.rebalance_ready import RebalanceReady


class SimpleKafkaLockRebalancer(ConsumerRebalanceListener):
    __REBALANCING_INTERVAL_CHECK = 0.3

    def __init__(self, kafka_python_multithreaded_consumer: RebalanceReady):
        self.kafka_python_multithreaded_consumer = kafka_python_multithreaded_consumer

    async def on_partitions_revoked(self, revoked):
        while not self.kafka_python_multithreaded_consumer.is_ready_for_rebalancing():
            await asyncio.sleep(self.__REBALANCING_INTERVAL_CHECK)

    async def on_partitions_assigned(self, assigned):
        pass
